#include <stdio.h>
#include <stdlib.h>

#include <time.h>
#include <signal.h>
#include <string.h>

#include <libunwind.h>

#define panic(args...) \
        do { fprintf (stderr, args); exit (EXIT_FAILURE); } while (0)

static int loop = 1;

static void unwind (void)
{
  unw_cursor_t cursor;
  unw_context_t uc;
  int ret, level = 0;
  unw_word_t ip, sp;

  /* Turn off caching.  */
  //unw_set_caching_policy (unw_local_addr_space, UNW_CACHE_NONE);

  unw_getcontext (&uc);

  if (unw_init_local (&cursor, &uc) < 0)
    panic ("unw_init_local() failed\n");

  do
    {
      ret = unw_step (&cursor);
      if (ret < 0)
        panic ("unw_step() failed\n");
      ++level;

      unw_get_reg (&cursor, UNW_REG_IP, &ip);
      unw_get_reg (&cursor, UNW_REG_SP, &sp);
      printf ("ip: %10p, sp: %10p\n", (void*) ip, (void*) sp);
    }
  while (ret > 0);

  printf ("Unwound %d levels\n", level);
}

static void sig_handler (int sig, siginfo_t *si, void *uc)
{
  unwind (); /* Note: Not async signal safe due to printf. */
  loop = 0;
}

static void setup_timer (void)
{
  struct sigaction act;
  struct sigevent sev;
  timer_t timerid;
  struct itimerspec its;

  memset (&its, 0, sizeof (its));

  /* establish a handler for the timer signal */
  act.sa_flags = SA_SIGINFO;
  act.sa_sigaction = sig_handler;

  if (sigaction(SIGRTMIN, &act, NULL) == -1)
    panic ("sigaction failed\n");

  /* create the timer */
  sev.sigev_notify = SIGEV_SIGNAL;
  sev.sigev_signo = SIGRTMIN;
  sev.sigev_value.sival_ptr = &timerid;
  if (timer_create(CLOCK_REALTIME, &sev, &timerid) == -1)
    panic ("timer_create\n");

  /* start the timer */
  its.it_value.tv_sec = 1; /* 1 second */
  if (timer_settime(timerid, 0, &its, NULL) == -1)
    panic ("timer_settime\n");
}

void lib2func (void)
{
  setup_timer ();

  while (loop);
}
